/**
 * Name:	SLAM Application
 * Author:	Johan Korsnes
 * Date:	2018
 *
 * Credits: The drivers in the project folder "drivers" are developed as part of
 *			the SLAM project. The code and FreeRTOS set-up is based on Nordic
 *			Semiconductor's example FreeRTOS project part of their SDK. The code
 *			and implementatino of Tracealyzer has been introduced as part of
 *			this project. The threads, queues and mutex have been set up as part
 *			of the SLAM project.
 *			
 *			Nordic Semiconductor's copyright notice can be found in
 *			ble_cleanup.h, as this is where the Bluetooth and example code has
 *			been moved to.
 */

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "nrf_sdh_freertos.h"
#include "app_timer.h"
#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "semphr.h"
#include "fds.h"
#include "nrf_drv_clock.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "nrf_drv_ppi.h"
#include "nrf_delay.h"

/* This main.c file originally contained almost 1300 lines of code, most if
 * being Bluetoother boiler plate code. To make the project more manageable
 * all this functionality was simply moved to the following header-file.
 * This was a quick-fix, and proper modularization should be performed.
 */
#include "ble_cleanup.h"

#include "display.h"
#include "motor.h"
#include "encoder.h"
#include "ir.h"
#include "servo.h"
#include "mag3110.h"
#include "microsd.h"

static TaskHandle_t handle_display_task,
					handle_user_task,
					handle_motor_task,
					handle_encoder_task,
					handle_ir_task,
					handle_servo_task,
					handle_mag_task,
					handle_microsd_task;
					
QueueHandle_t queue_microsd;

SemaphoreHandle_t mutex_spi;

// all SPI driver interaction occurs within mutex, so can safely use global bool
bool shared_SPI_init = false;

/**@brief A function which is hooked to idle task.
 * @note Idle hook must be enabled in FreeRTOS configuration (configUSE_IDLE_HOOK).
 */
void vApplicationIdleHook( void )
{
#if NRF_LOG_ENABLED
     // vTaskResume(m_logger_thread);
#endif
}

/**@brief Function for initializing the clock.
 */
static void clock_init(void)
{
    ret_code_t err_code = nrf_drv_clock_init();
    APP_ERROR_CHECK(err_code);
}

/**@brief Motor task controlling.
 *
 * @details Initializes the motor driver and performs motor operations.
 */
static void motor_task(void* arg) {
	motor_init();
	
	microsd_write_operation_t write = {
		.filename = "motor",
		.content = ""
	};
	
	for ( ;; ) {
		NRF_LOG_INFO("Starting motor!\n");
		write.content = "moving forward";
//		xQueueSendToBack(queue_microsd, &write, portMAX_DELAY);
		motor_forward(20);
		vTaskDelay(2000);
		write.content = "moving stopping";
//		xQueueSendToBack(queue_microsd, &write, portMAX_DELAY);
		motor_stop();
		vTaskDelay(2000);
	}
}

long encoder_left = 0;
long encoder_right = 0;

/**@brief Encoder sensor callback for left wheel.
 *
 * @details Used to update encoder count for interrupt initialized encoders.
 */
void encoder_cb_left() {
	encoder_left++;
};

/**@brief Encoder sensor callback for right wheel.
 *
 * @details Used to update encoder count for interrupt initialized encoders.
 */
void encoder_cb_right() {
	encoder_right++;
};

/**@brief Encoder sensor task.
 *
 * @details Initializes the encoder in interrupt mode, and logs the encoder
 *			values with an interval.
 */
static void encoder_task(void* arg) {
	// consider using PPI for one of them? Yes.
	encoder_init_int(ENCODER_LEFT, encoder_cb_left);
	encoder_init_int(ENCODER_RIGHT, encoder_cb_right);
	for ( ;; ) {
		NRF_LOG_INFO("\nEncoder left: %d, Encoder RIght: %d\n",
					 encoder_left, encoder_right);
		vTaskDelay(2000);
	}
}

long ir_val = 0;


/**@brief IR sensor callback.
 *
 * @details Updates the ir-reading value.
 */
void ir_cb(uint16_t val) {
	ir_val = val;
}

/**@brief IR sensor task.
 *
 * @details Task for obtaining ADC measurements from the IR sensor.
 */
static void ir_task(void* arg) {
	ir_init(IR_RESOLUTION_10);
	
	for ( ;; ) {
		ir_read(IR_SENSOR_4, ir_cb);
		vTaskDelay(1000);
		NRF_LOG_INFO("\nIR value: %d", ir_val);
	}
}

/**@brief Servo task.
 *
 * @details Task for interacting with the sensor-tower servo.
 */
static void servo_task(void* arg) {
	servo_init();
	
	for ( ;; ) {
		NRF_LOG_INFO("\nservo!");
		servo_position(10);
		vTaskDelay(1000);
		servo_position(20);
		vTaskDelay(1000);
	}
}

/**@brief Magnetometer task
 *
 * @details Task to initialize, calibrate and read heading from magnetometer.
 *			Note that the calibration requires the full physical rotation of the
 *			control system, and has therefore been commented out by default.
 */
static void mag_task(void* arg) {
	mag_init(MAG_OS_128);
	// mag_calibrate();
	
	for ( ;; ) {
		// float heading = mag_heading()
		// NRF_LOG_INFO("Heading:" NRF_LOG_FLOAT_MARKER "\r\n", NRF_LOG_FLOAT(mag_heading()));
		vTaskDelay(1000);
	}
}

void display_task(void* arg) {
	display_init();
	
	display_operation_t display_operation;
	microsd_write_operation_t write = {
		.filename = "DISP",
		.content = "What did I just draw"
	};
	
	// uncomment lines below for demo
//	nrf_gfx_display(&m_nrf_lcd);
//	for (;;);
	
	for( ;; ) {
		xQueueReceive(queue_display, &display_operation, portMAX_DELAY);
		xSemaphoreTake(mutex_spi, portMAX_DELAY);
		switch (display_operation.operation) {
			case DISPLAY_TEXT:
				APP_ERROR_CHECK(nrf_gfx_print(&m_nrf_lcd, &display_operation.element.text.text_position,
				1, display_operation.element.text.p_string, &orkney_8ptFontInfo, false));
				write.content = "Text drawn.";
				break;
			case DISPLAY_POINT:
				nrf_gfx_point_draw(&m_nrf_lcd, &display_operation.element.point.point, 1);
				write.content = "Point drawn.";
				break;
			case DISPLAY_LINE:
				APP_ERROR_CHECK(nrf_gfx_line_draw(
					&m_nrf_lcd, &display_operation.element.line.line, 1));
					write.content = "Line drawn.";
				break;
			case DISPLAY_CIRCLE:
				APP_ERROR_CHECK(nrf_gfx_circle_draw(
					&m_nrf_lcd, &display_operation.element.circle.circle,
					1, display_operation.element.circle.fill));
					write.content = "Circle.";
				break;
			case DISPLAY_RECTANGLE:
				APP_ERROR_CHECK(nrf_gfx_rect_draw(
					&m_nrf_lcd, &display_operation.element.rectangle.rectangle,
					display_operation.element.rectangle.thickness, 1,
					display_operation.element.rectangle.fill));
					write.content = "Rectangle drawn.";
				break;
			case DISPLAY_LOG:
				// print log buffer
				break; // not yet implemented
			case DISPLAY_CLEAR:
				nrf_gfx_screen_fill(&m_nrf_lcd, 0);
				break;
			
			default:
				;
		}
		nrf_gfx_display(&m_nrf_lcd);
		xSemaphoreGive(mutex_spi);
		xQueueSendToBack(queue_microsd, &write, portMAX_DELAY);
	}
}

/**@brief User task
 *
 * @details Task for miscellaneous operations. Currently only adding display ops
 *			to the display_task's queue.
 */
static void user_task(void* arg) {
	microsd_write_operation_t write;
	write.filename = "USR";
	write.content = "write";
	for ( ;; ) {
		display_text(5, 1, "test of clear");
		vTaskDelay(1000);
		display_circle(45,35,20,false);
		vTaskDelay(1000);
		display_rectangle(10, 30, 15, 10, 2, true);
		vTaskDelay(1000);
		display_point(2, 2);
		vTaskDelay(1000);
		display_line(1, 32, 125, 32, 2);
		vTaskDelay(1000);
		display_clear();
		vTaskDelay(1000);
		xQueueSendToBack(queue_microsd, &write, portMAX_DELAY);
		vTaskDelay(2000);
		write.filename = "USR";
		write.content = "I've been drawing.";
		vTaskDelay(2500);
		display_text(5, 1, "restarting loop");
	}
}

/**@brief Task for handling microSD operations.
 *
 * @details Initializes the microSD driver, and performs write requests.
 */
static void microsd_task(void* arg) {
	microsd_write_operation_t write_operation;
	for ( ;; ) {
		xQueueReceive(queue_microsd, &write_operation, portMAX_DELAY);
//		xSemaphoreTake(mutex_spi, portMAX_DELAY);
		NRF_LOG_INFO("writing %s.", write_operation.content);
		microsd_write(write_operation.filename, write_operation.content);
//		xSemaphoreGive(mutex_spi);
	}
}

/**@brief Function for application main entry.
 */
int main(void)
{
    bool erase_bonds;
    log_init();
    clock_init();

    // Do not start any interrupt that uses system functions before system initialisation.
    // The best solution is to start the OS before any other initalisation.

	/* Deferred logging has been disabled for the SLAM application
	#if NRF_LOG_ENABLED
		if (pdPASS != xTaskCreate(logger_thread, "LOGGER", 256, NULL, 1, &m_logger_thread))
			APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
	#endif
	*/

    SCB->SCR |= SCB_SCR_SLEEPDEEP_Msk;
    ble_stack_init();
	
	/**
	* The reset-reason functions are extremely useful as they help you rule out
	* issues with power supply without having to resort to a oscilloscope.
	*/
	#if !NRF_LOG_ENABLED || !NRF_LOG_BACKEND_RTT_ENABLED
//	char const * error_msg;
//	vTraceEnable(TRC_INIT);
//	error_msg = xTraceGetLastError();
	#endif
	
    // Initialize modules.
    timers_init();
    gap_params_init();
    gatt_init();
    advertising_init();
    services_init();
    sensor_simulator_init();
    conn_params_init();
    peer_manager_init();
    application_timers_start();
	
	/**
	* The reset-reason functions are extremely useful as they help you rule out
	* issues with power supply without having to resort to a oscilloscope.
	*/
	//sd_power_reset_reason_get(&reset_reason);
	//sd_power_reset_reason_clr(0xFFFFFFFF);
	
	/**
	* The reset-reason functions are extremely useful as they help you rule out
	* issues with power supply without having to resort to a oscilloscope.
	*/
	// RecorderDataPtr
	#if !NRF_LOG_ENABLED || !NRF_LOG_BACKEND_RTT_ENABLED
	vTraceEnable(TRC_START);
	error_msg = xTraceGetLastError();
	#endif
	
	queue_display = xQueueCreate(5, sizeof(display_operation_t));
	queue_microsd = xQueueCreate(5, sizeof(microsd_write_operation_t));
	
	mutex_spi = xSemaphoreCreateMutex();	

	if (pdPASS != xTaskCreate(display_task, "DISP", 256, NULL, 1, handle_display_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
//
	if (pdPASS != xTaskCreate(user_task, "USER", 128, NULL, 1, handle_user_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
//
	if (pdPASS != xTaskCreate(motor_task, "MOTOR", 128, NULL, 1, handle_motor_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
//	
	if (pdPASS != xTaskCreate(encoder_task, "ENC", 128, NULL, 1, handle_encoder_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
	
	if (pdPASS != xTaskCreate(ir_task, "IR", 128, NULL, 1, handle_ir_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
	
	if (pdPASS != xTaskCreate(servo_task, "SERVO", 128, NULL, 1, handle_servo_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
	
	// mag_task is unstable
//	if (pdPASS != xTaskCreate(mag_task, "MAG", 64, NULL, 1, handle_mag_task))
//		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
//		
	if (pdPASS != xTaskCreate(microsd_task, "SD", 256, NULL, 1, handle_microsd_task))
		APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
//	
//    // Create a FreeRTOS task for the BLE stack.
//    // The task will run advertising_start() before entering its loop.
    nrf_sdh_freertos_init(advertising_start, &erase_bonds);

    NRF_LOG_INFO("\nInitialization done. SLAM application now starting.\n.");
	
    vTaskStartScheduler();

    for ( ;; )
    {
		/**
		* vTaskStartSchedule returns only if the system failed to allocate heap
		* memory. Either reduce heap memory usage, or increase the allocation
		* of heap memory in FreeRTOSConfig.h via configTOTAL_HEAP_SIZE
		*/
        APP_ERROR_HANDLER(NRF_ERROR_FORBIDDEN);
    }
}
